#include "fdc.h"


#define	RSTDELAY			86		// FDCZbg̕Wx(us)

#define	SCANRESULT			(sizeof(FD_TIMED_SCAN_RESULT)+sizeof(FD_TIMED_ID_HEADER)*MAX_SECTOR)


////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
FDC::FDC( void )
{
	h      = NULL;
	Retry  = 10;
	ZeroMemory( &CmdRes, sizeof(CmdRes) );
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
FDC::~FDC( void )
{
	if( h ) CloseHandle( h );
}


////////////////////////////////////////////////////////////////
// ep[^擾Eݒ
////////////////////////////////////////////////////////////////
BYTE FDC::ST0( void ){ return CmdRes.st0; }		// UgXe[^X擾
BYTE FDC::ST1( void ){ return CmdRes.st1; }
BYTE FDC::ST2( void ){ return CmdRes.st2; }

void FDC::SetRetry( int rt ){ Retry = rt; }		// gC񐔐ݒ



////////////////////////////////////////////////////////////////
// Raw command
////////////////////////////////////////////////////////////////
bool FDC::SendRawCommand( DWORD command, LPVOID ibuf, DWORD isize, LPVOID obuf, DWORD osize )
{
	int rt = 0;
	DWORD ret;
	
	if( !h ) return false;
	
	while( rt++ < Retry ){
		if( DeviceIoControl( h, command, ibuf, isize, obuf, osize, &ret, NULL ) )
			return true;
	}
	return false;
}

bool FDC::CmdRecalibrate( void )
{
	return SendRawCommand( IOCTL_FDCMD_RECALIBRATE, NULL, 0, &Ist, sizeof(Ist) );
}

bool FDC::CmdSeek( BYTE pcyl )
{
	FD_SEEK_PARAMS sep = { pcyl, 0 };
	return SendRawCommand( IOCTL_FDCMD_SEEK, &sep, sizeof(sep), &Ist, sizeof(Ist) );
}

bool FDC::CmdReadID( BYTE phead, BYTE flag )
{
	FD_READ_ID_PARAMS rip = { flag, phead };
	return SendRawCommand( IOCTL_FDCMD_READ_ID, &rip, sizeof(rip), &CmdRes, sizeof(CmdRes) );
}

bool FDC::CmdReadData( BYTE phead, BYTE cyl, BYTE head, BYTE sector, BYTE size, BYTE *pv, BYTE flag )
{
	FD_READ_WRITE_PARAMS rwp = { flag, phead, cyl, head, sector, size > 8 ? (BYTE)8 : size, (BYTE)(sector+1), (BYTE)(7<<size), 0xff };
	bool ret = SendRawCommand( IOCTL_FDCMD_READ_DATA, &rwp, sizeof(rwp), pv, 128<<rwp.size );
	GetResult();
	return ret;
}

bool FDC::CmdReadDeletedData( BYTE phead, BYTE cyl, BYTE head, BYTE sector, BYTE size, BYTE *pv, BYTE flag )
{
	FD_READ_WRITE_PARAMS rwp = { flag, phead, cyl, head, sector, size > 8 ? (BYTE)8 : size, (BYTE)(sector+1), (BYTE)(7<<size), 0xff };
	bool ret = SendRawCommand( IOCTL_FDCMD_READ_DELETED_DATA, &rwp, sizeof(rwp), pv, 128<<rwp.size );
	GetResult();
	return ret;
}

bool FDC::CmdWriteData( BYTE phead, BYTE cyl, BYTE head, BYTE sector, BYTE size, BYTE *pv, BYTE flag )
{
	FD_READ_WRITE_PARAMS rwp = { flag, phead, cyl, head, sector, size > 8 ? (BYTE)8 : size, (BYTE)(sector+1), (BYTE)(7<<size), 0xff };
	bool ret = SendRawCommand( IOCTL_FDCMD_WRITE_DATA, &rwp, sizeof(rwp), pv, 128<<rwp.size );
	GetResult();
	return ret;
}

bool FDC::CmdWriteDeletedData( BYTE phead, BYTE cyl, BYTE head, BYTE sector, BYTE size, BYTE *pv, BYTE flag )
{
	FD_READ_WRITE_PARAMS rwp = { flag, phead, cyl, head, sector, size > 8 ? (BYTE)8 : size, (BYTE)(sector+1), (BYTE)(7<<size), 0xff };
	bool ret = SendRawCommand( IOCTL_FDCMD_WRITE_DELETED_DATA, &rwp, sizeof(rwp), pv, 128<<rwp.size );
	GetResult();
	return ret;
}

bool FDC::CmdFormat( PFD_FORMAT_PARAMS pfp )
{
	return SendRawCommand( IOCTL_FDCMD_FORMAT_TRACK, pfp, sizeof(FD_FORMAT_PARAMS)+sizeof(PFD_ID_HEADER)*pfp->sectors, NULL, 0 );
}

bool FDC::CmdSpecify( BYTE srt, BYTE hut, BYTE hlt )
{
	FD_SPECIFY_PARAMS sp = { (BYTE)((srt<<4)|(hut&0x0f)), (BYTE)((hlt<<1)|1) };
	return SendRawCommand( IOCTL_FDCMD_SPECIFY, &sp, sizeof(sp), &CmdRes, sizeof(CmdRes) );
}


bool FDC::Reset( void )
{
	return SendRawCommand( IOCTL_FD_RESET, NULL, 0, NULL, 0 );
}

bool FDC::SetDataRate( BYTE rate )
{
	return SendRawCommand( IOCTL_FD_SET_DATA_RATE, &rate, sizeof(rate), NULL, 0 );
}

int FDC::ScanTrack( BYTE head, SECINF *sec, BYTE flag )
{
	BYTE scanbuf[SCANRESULT];
	FD_SCAN_PARAMS scp = { flag, head };
	
	if( !sec ) return -1;
	
	PFD_TIMED_SCAN_RESULT psr = (PFD_TIMED_SCAN_RESULT)scanbuf;
	PFD_TIMED_ID_HEADER ph    = psr->Header;
	
	ZeroMemory( scanbuf, sizeof(scanbuf) );
	if( !SendRawCommand( IOCTL_FD_TIMED_SCAN_TRACK, &scp, sizeof(scp), (PFD_TIMED_SCAN_RESULT)scanbuf, sizeof(scanbuf) ) )
		return -1;
	
	SECINF *cs = sec;
	for( int i=0; i < psr->count; i++,cs++,ph++ ){
		cs->c       = ph->cyl;
		cs->h       = ph->head;
		cs->r       = ph->sector;
		cs->n       = ph->size;
		cs->reltime = ph->reltime;
		cs->density = flag;
		
		SECINF *ss = sec;
		int soff = 0;
		while( ss != cs ){
			if( (ss->c == cs->c) && (ss->h == cs->h) &&	(ss->r == cs->r) && (ss->n == cs->n) )
				ss->secoff = ++soff;
			ss++;
		}
		if( soff ) cs->secoff = ++soff;
	}
	return psr->count;
}

bool FDC::GetResult( void )
{
	return SendRawCommand( IOCTL_FD_GET_RESULT, NULL, 0, &CmdRes, sizeof(CmdRes) );
}

bool FDC::SetSectorOffset( BYTE offset )
{
	FD_SECTOR_OFFSET_PARAMS sop = { offset };
	return SendRawCommand( IOCTL_FD_SET_SECTOR_OFFSET, &sop, sizeof(sop), NULL, 0 );
}

bool FDC::SetShortWrite( DWORD size, int delay )
{
	FD_SHORT_WRITE_PARAMS swp = { size, (DWORD)(RSTDELAY+delay) };
	return SendRawCommand( IOCTL_FD_SET_SHORT_WRITE, &swp, sizeof(swp), NULL, 0 );
}

bool FDC::CheckDrive( void )
{
	return SendRawCommand( IOCTL_FD_CHECK_DISK, NULL, 0, &CmdRes, sizeof(CmdRes) );
}




////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
DWORD FDC::GetDriverVersion( void )
{
	DWORD ver = 0;
	DWORD ret;
	
	HANDLE dh = CreateFile( "\\\\.\\fdrawcmd", GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
	if( dh != INVALID_HANDLE_VALUE ){
		DeviceIoControl( dh, IOCTL_FDRAWCMD_GET_VERSION, NULL, 0, &ver, sizeof(ver), &ret, NULL );
		CloseHandle( dh );
	}
	return ver;
}





////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
bool FDC::OpenFDD( int drv, BYTE rate )
{
	char str[32];
	
	wsprintf( str, "\\\\.\\fdraw%u", drv&1 );
	
	h = CreateFile( str, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
    return ( h != INVALID_HANDLE_VALUE && SetDataRate( rate ) ) ? true : false;
}


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
int FDC::Init( int dnum, BYTE rate )
{
	DWORD ver = GetDriverVersion();
	
	// fdrawcmd.sys ݃`FbN
	if( ver == 0 ) return 1;
	
	// fdrawcmd.sys o[W`FbN
	if( HIWORD(ver) != HIWORD(FDRAWCMD_VERSION) ) return 2;
	
	if( !OpenFDD( dnum, rate ) ) return 3;
	
	if( !Reset() ){
		CloseHandle( h );
		h = NULL;
		return 4;
	}
	
	CheckDrive();
	if( GetLastError() == ERROR_NO_MEDIA_IN_DRIVE ) return 5;
	
	CmdRecalibrate();
	CmdRecalibrate();
	
	return 0;
}
